from luz.datasets import *
from luz.events import *
from luz.flags import *
from luz.handlers import *
from luz.learner import *
from luz.loggers import *
from luz.model import *
from luz.modules import *
from luz.scorers import *
from luz.transforms import *
from luz.tuners import *
from luz.utils import *



from __future__ import annotations
from typing import Any, Iterable, Iterator, Optional, Union

from abc import ABC, abstractmethod
import ast
import collections
import copy
import itertools
import json
import luz
import numpy as np
import operator
import pathlib
import re
import torch

__all__ = ["Tuner", "BayesianTuner", "GridSearch", "RandomSearch"]

Device = Union[str, torch.device]


class Experiment:
    def __init__(self, **kwargs: Any) -> None:
        self.kwargs = kwargs

    def __getattr__(self, key: str) -> Any:
        return self.kwargs[key]

    def encode(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, torch.nn.Module):
            if len(obj._modules) > 0:
                return obj._modules
            else:
                return repr(obj)

    def json(self, score: float, model: luz.Module) -> str:
        d = {"hyperparameters": self.kwargs, "score": score, "model": model}
        return json.dumps(d, default=self.encode, indent=4)

    def __str__(self):
        return json.dumps(
            {"hyperparameters": self.kwargs}, default=self.encode, indent=4
        )


class Sample:
    def __init__(self, lower: Union[int, float], upper: Union[int, float]) -> None:
        self.lower = lower
        self.upper = upper


class Choose:
    def __init__(self, choices: Iterable[Any]) -> None:
        self.choices = tuple(choices)


class Pin:
    def __init__(self, equation: str) -> None:
        self.equation = equation

    def __call__(self, **kwargs: Any) -> Any:
        # NOTE: this sorting should ensure that longer
        # variable names are substituted first, which
        # should prevent partial variable name clobbering
        pattern = "|".join(sorted(kwargs, key=len)[::-1])

        def replacement(m):
            return str(kwargs[re.escape(m.group(0))])

        expression = re.sub(pattern=pattern, repl=replacement, string=self.equation)

        # parse pin equation string into operators and
        # evaluate operators to obtain the pin value
        return _evaluate_expression(expression)


class Conditional:
    def __init__(self, condition: str, if_true: Any, if_false: Any) -> None:
        self.pin = Pin(condition)
        self.if_true = if_true
        self.if_false = if_false

    def __call__(self, **kwargs: Any) -> Any:
        if self.pin(**kwargs):
            return self.if_true
        return self.if_false


class Tuner(ABC):
    def __init__(
        self,
        num_iterations: int,
        seed: Optional[int] = 0,
        json_dir: Optional[str] = None,
    ) -> None:
        """Hyperparameter tuning algorithm.

        Parameters
        ----------
        num_iterations
            Number of tuning iterations.
        seed
            Random seed for consistency across tuning iterations; by default 0.
        """
        self.num_iterations = num_iterations
        self.seed = seed
        self.json_dir = json_dir

        self.hyperparameters = []

        self.experiment = None

        self.scores = collections.defaultdict(list)
        self.best_model = None

    @abstractmethod
    def sample_hyperparameters(self, **samples: Sample) -> Union[int, float]:
        """Sample hyperparameter value.

        Parameters
        ----------
        lower
            Hyperparameter lower bound.
        upper
            Hyperparameter upper bound.

        Returns
        -------
        Union[int, float]
            Sampled hyperparameter value.
        """
        pass

    @abstractmethod
    def choose_hyperparameters(self, **choices: Choose) -> Any:
        """Choose hyperparameter value.

        Parameters
        ----------
        choices
            Possible hyperparameter values.

        Returns
        -------
        Any
            Chosen hyperparameter value.
        """
        pass

    def sample(self, lower: Union[int, float], upper: Union[int, float]) -> Sample:
        s = Sample(lower, upper)
        self.hyperparameters.append(s)

        return s

    def choose(self, *choices: Any) -> Choose:
        c = Choose(choices=choices)
        self.hyperparameters.append(c)

        return c

    def pin(self, equation: str) -> Pin:
        p = Pin(equation)
        self.hyperparameters.append(p)

        return p

    def conditional(self, condition: str, if_true: Any, if_false: Any) -> Conditional:
        c = Conditional(condition, if_true, if_false)
        self.hyperparameters.append(c)

        return c

    # NOTE: any pins or conditionals have to come after the variables they reference!
    def tune(self, **kwargs: Union[Choose, Conditional, Pin, Sample]) -> Iterator[Any]:
        """Tune hyperparameters. Wrap code to be run on each tuning iteration.

        Yields
        -------
        Iterator[Any]
            Hyperparameter values.
        """
        for i in range(self.num_iterations):
            self.current_iteration = i

            try:
                sample_hps = {k: v for k, v in kwargs.items() if isinstance(v, Sample)}
                samples = self.sample_hyperparameters(**sample_hps)
            except NotImplementedError:
                pass

            try:
                choice_hps = {k: v for k, v in kwargs.items() if isinstance(v, Choose)}
                choices = self.choose_hyperparameters(**choice_hps)
            except NotImplementedError:
                pass

            d = {}

            for k, v in kwargs.items():
                if k in sample_hps:
                    d[k] = samples[k]
                elif k in choice_hps:
                    d[k] = choices[k]
                else:
                    d[k] = v(**d)

            self.experiment = Experiment(**d)

            if self.experiment in self.scores:
                self.scores[self.experiment].append(self.scores[k][-1])
            else:
                # seed to ensure reproducibility in each iteration of the tuning loop
                with luz.temporary_seed(self.seed):
                    yield self.experiment

    def score(
        self, learner: luz.Learner, dataset: luz.Dataset, device: Device
    ) -> luz.Score:
        """
        Learn a model and estimate its future performance based on a given dataset.

        Parameters
        ----------
        learner
            Learning algorithm to be scored.
        dataset
            Dataset used to learn and score a model.
        device
            Device to use for learning, by default "cpu".

        Returns
        -------
        luz.Score
            Learned model and estimated performance.
        """
        model, score = learner.score(dataset, device)

        if self.best_model is None or score < self.best_score:
            self.best_model = copy.deepcopy(model)

        self.scores[self.experiment].append(score)

        if self.json_dir is not None:
            p = pathlib.Path(self.json_dir, f"{self.current_iteration}.json")
            with open(str(p), "w") as f:
                f.write(self.experiment.json(score, model))

        return luz.Score(model, score)

    @property
    def best_hyperparameters(self):
        mean_scores = {k: sum(v) / len(v) for k, v in self.scores.items()}
        try:
            k = min(mean_scores, key=mean_scores.get)
            return k.kwargs
        except ValueError:
            return None

    @property
    def best_score(self):
        mean_scores = {k: sum(v) / len(v) for k, v in self.scores.items()}
        try:
            k = min(mean_scores, key=mean_scores.get)
            return mean_scores[k]
        except ValueError:
            return None


class BayesianTuner(Tuner):
    def sample_hyperparameter(
        self, lower: Union[int, float], upper: Union[int, float]
    ) -> Union[int, float]:
        raise NotImplementedError

    def choose_hyperparameter(self, choices: Iterable[Any]) -> Any:
        raise NotImplementedError


class GridSearch(Tuner):
    def __init__(self, seed_loop: Optional[bool] = False) -> None:
        """Hyperparameter tuning algorithm.

        Parameters
        ----------
        seed_loop
            If True seed each iteration, by default False.
        """
        super().__init__(0, seed_loop)
        self.grid = None

    def tune(self, **kwargs: Union[Choose, Conditional, Pin, Sample]) -> Iterator[Any]:
        choices = [hp for hp in self.hyperparameters if isinstance(hp, Choose)]
        self.grid = itertools.product(*[c.choices for c in choices])
        self.num_iterations = np.prod([len(c.choices) for c in choices])

        yield from super().tune(**kwargs)

    def sample_hyperparameters(self, **samples: Sample) -> Union[int, float]:
        raise NotImplementedError

    def choose_hyperparameters(self, **choices: Choose) -> Any:
        return dict(zip(choices.keys(), next(self.grid)))


class RandomSearch(Tuner):
    def sample_hyperparameters(self, **samples: Sample) -> Union[int, float]:
        """Sample hyperparameter value.

        Parameters
        ----------
        lower
            Hyperparameter lower bound.
        upper
            Hyperparameter upper bound.

        Returns
        -------
        Union[int, float]
            Sampled hyperparameter value from the range `[lower, upper)`.
        """
        d = {}
        for k, s in samples.items():
            (datatype,) = set((type(s.lower), type(s.upper)))
            if datatype == int:
                d[k] = np.random.randint(low=s.lower, high=s.upper)
            elif datatype == float:
                d[k] = np.random.uniform(low=s.lower, high=s.upper)
            else:
                d[k] = None

        return d

    def choose_hyperparameters(self, **choices: Choose) -> Any:
        """Choose hyperparameter value.

        Parameters
        ----------
        choices
            Possible hyperparameter values.

        Returns
        -------
        Any
            Chosen hyperparameter value.
        """
        return {k: np.random.choice(a=c.choices) for k, c in choices.items()}


# from https://stackoverflow.com/a/9558001

ops = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.Div: operator.truediv,
    ast.Pow: operator.pow,
    ast.BitXor: operator.xor,
    ast.USub: operator.neg,
    ast.FloorDiv: operator.floordiv,
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    ast.Is: operator.is_,
    ast.IsNot: operator.is_not,
}


def _evaluate_expression(expression: str) -> Any:
    node = ast.parse(expression, mode="eval").body

    return _evaluate_operators(node)


def _evaluate_operators(node) -> Any:
    # FIXME: type annotate `node`
    # <number>
    if isinstance(node, ast.Num):
        return node.n
    # <left> <operator> <right>
    elif isinstance(node, ast.BinOp):
        return ops[type(node.op)](
            _evaluate_operators(node.left),
            _evaluate_operators(node.right),
        )
    # <operator> <operand> e.g., -1
    elif isinstance(node, ast.UnaryOp):
        return ops[type(node.op)](_evaluate_operators(node.operand))
    elif isinstance(node, ast.Compare):
        # from https://github.com/danthedeckie/simpleeval/blob/master/simpleeval.py
        right = _evaluate_operators(node.left)
        to_return = True
        for operation, comp in zip(node.ops, node.comparators):
            if not to_return:
                break
            left = right
            right = _evaluate_operators(comp)
            to_return = ops[type(operation)](left, right)
        return to_return
    else:
        raise TypeError(node)



from __future__ import annotations
from typing import Any, Iterable, Optional, Union

from abc import ABC, abstractmethod
import collections
import luz
import torch

__all__ = [
    "Argmax",
    "Compose",
    "Identity",
    "Lookup",
    "Normalize",
    "PowerSeries",
    "Reshape",
    "Squeeze",
    "Standardize",
    "TensorTransform",
    "Transform",
    "Transpose",
    "Unsqueeze",
    "ZeroMeanPerTensor",
]


class Transform:
    def __init__(self, **transforms: TensorTransform) -> None:
        self.transforms = collections.defaultdict(Identity, transforms)

    def __call__(self, data: luz.Data) -> luz.Data:
        """Transform data.

        Parameters
        ----------
        data
            Input data.

        Returns
        -------
        luz.Data
            Output data.
        """
        kw = {k: self.transforms[k](data[k]) for k in data.keys}
        return luz.Data(**kw)


class TensorTransform(ABC):
    @abstractmethod
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        "Transform which is applied to an input tensor."

    def inverse(self) -> TensorTransform:
        raise NotImplementedError


# INVERTIBLE


class Compose(TensorTransform):
    def __init__(self, *transforms: Iterable[TensorTransform]) -> None:
        self.transforms = transforms

    def __call__(self, x: Any) -> Any:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        for transform in self.transforms:
            x = transform(x)
        return x

    def inverse(self) -> TensorTransform:
        inverse_transforms = [t.inverse() for t in self.transforms][::-1]
        return Compose(*inverse_transforms)


class Squeeze(TensorTransform):
    def __init__(self, dim: Optional[int]) -> None:
        """Squeeze tensor.

        Parameters
        ----------
        dim
            Dimension to be squeezed.
        """
        super().__init__()
        self.dim = dim

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Squeezed output tensor.
        """
        return x.squeeze(dim=self.dim)

    def inverse(self) -> TensorTransform:
        return Unsqueeze(self.dim)


class Unsqueeze(TensorTransform):
    def __init__(self, dim: Optional[int]) -> None:
        """Unsqueeze tensor.

        Parameters
        ----------
        dim
            Dimension to be unsqueezed.
        """
        super().__init__()
        self.dim = dim

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Unsqueezed output tensor.
        """
        return x.unsqueeze(dim=self.dim)

    def inverse(self) -> TensorTransform:
        return Squeeze(self.dim)


class Identity(TensorTransform):
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor, same as the input tensor.
        """
        return x

    def inverse(self) -> TensorTransform:
        return Identity()


class Standardize(TensorTransform):
    def __init__(self, mean: torch.Tensor, std: Optional[torch.Tensor] = None) -> None:
        self.mean = mean
        self.std = std

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Standardized output tensor.
        """
        return (x - self.mean) / self.std

    def inverse(self) -> TensorTransform:
        return Standardize(-self.mean / self.std, 1 / self.std)


class Transpose(TensorTransform):
    def __init__(self, dim0: int, dim1: int) -> None:
        self.dim0 = dim0
        self.dim1 = dim1

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Transposed output tensor.
        """
        return torch.transpose(x, self.dim0, self.dim1)

    def inverse(self) -> TensorTransform:
        return Transpose(self.dim0, self.dim1)


# NON-INVERTIBLE


class Argmax(TensorTransform):
    def __init__(
        self, dim: Optional[int] = None, keepdim: Optional[bool] = False
    ) -> None:
        self.dim = dim
        self.keepdim = keepdim

    def __call__(self, x: torch.Tensor) -> torch.LongTensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return x.argmax(dim=self.dim, keepdim=self.keepdim)


class Lookup(TensorTransform):
    def __init__(self, lookup_dict) -> None:
        self.lookup_dict = lookup_dict

    def __call__(self, x) -> Any:
        """Transform tensor.

        Parameters
        ----------
        x
            Input.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return self.lookup_dict[x]


class Normalize(TensorTransform):
    def __init__(self, p: Union[int, str], *args, **kwargs) -> None:
        self.p = p
        self.args = args
        self.kwargs = kwargs

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Normalized output tensor.
        """
        return x / torch.linalg.norm(x, self.p, *self.args, **self.kwargs)


class PowerSeries(TensorTransform):
    def __init__(self, degree: int, dim: Optional[int] = -1) -> None:
        self.degree = degree
        self.dim = dim

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        tensors = tuple(x ** k for k in range(1, self.degree + 1))
        try:
            return torch.cat(tensors=tensors, dim=self.dim)
        except RuntimeError:
            return torch.stack(tensors=tensors, dim=self.dim)


class Reshape(TensorTransform):
    def __init__(self, out_shape: Iterable[int]) -> None:
        """Reshape tensor.

        Parameters
        ----------
        out_shape
            Desired output shape.
        """
        self.shape = tuple(out_shape)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Reshaped output tensor.
        """
        return x.view(self.shape)


class ZeroMeanPerTensor(TensorTransform):
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return x - torch.mean(x)



from __future__ import annotations
from typing import Callable, Optional, Union

import contextlib
import luz
import pathlib
import torch

__all__ = ["Model"]

Device = Union[str, torch.device]
Criterion = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]


class Model(torch.nn.Module):
    def run_batch(
        self,
        data: torch.Tensor,
        target: torch.Tensor,
        criterion: Optional[Criterion] = None,
        optimizer: Optional[torch.optim.Optimizer] = None,
    ) -> float:
        """Run training algorithm on a single batch.

        Parameters
        ----------
        dataset
            Batch of training data.
        target
            Target tensor.
        criterion
            Training criterion, by default None.
        optimizer
            Training optimizer, by default None.

        Returns
        -------
        torch.Tensor
            Model output.
        float
            Batch loss.
        """
        output = self(data)
        batch_loss = criterion(output, target)

        if optimizer is not None:
            self.backward(batch_loss)
            self.optimizer_step(optimizer)

        return output, batch_loss

    def run_train_batch(self, data, target, criterion, optimizer):
        return self.run_batch(data, target, criterion, optimizer)

    def run_validate_batch(self, data, target, criterion):
        return self.run_batch(data, target, criterion)

    def run_test_batch(self, data, target, criterion):
        return self.run_batch(data, target, criterion)

    def backward(self, loss: torch.Tensor) -> None:
        """Backpropagate loss.

        Parameters
        ----------
        loss
            Loss tensor.
        """
        loss.backward()

    def optimizer_step(self, optimizer: torch.optim.Optimizer) -> None:
        """Step training optimizer.

        Parameters
        ----------
        optimizer
            Training optimizer.
        """
        optimizer.step()
        optimizer.zero_grad()

    def get_input(self, batch: luz.Data) -> torch.Tensor:
        """Get input from batched data.

        Parameters
        ----------
        batch
            Batched data.

        Returns
        -------
        torch.Tensor
            Input tensor.
        """
        return batch.x

    def get_target(self, batch: luz.Data) -> Optional[torch.Tensor]:
        """Get target from batched data.

        Parameters
        ----------
        batch
            Batched data.

        Returns
        -------
        Optional[torch.Tensor]
            Target tensor.
        """
        return batch.y

    # NON-CONFIGURABLE METHODS BELOW

    @property
    def num_parameters(self) -> int:
        """Number of trainable parameters.

        Returns
        -------
        int
            Number of trainable parameters.
        """
        return sum(p.numel() for p in self.parameters() if p.requires_grad)

    def save(self, path: Union[str, pathlib.Path]) -> None:
        torch.save(self.state_dict(), path)
        #     {"model": self.state_dict(), "trainer": self.trainer.state_dict()}, path
        # )

    def load(self, path: Union[str, pathlib.Path]):
        state_dict = torch.load(path)
        self.load_state_dict(state_dict)  # ["model"])

        # self.trainer = luz.Trainer()
        # self.trainer.load_state_dict(state_dict["trainer"])

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """Compute forward pass in eval mode.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        with self.eval():
            if hasattr(self, "transform"):
                return self.transform(self.__call__(x))
            return self.__call__(x)

    @contextlib.contextmanager
    def eval(self, no_grad: Optional[bool] = True) -> None:
        """Context manager to operate in eval mode.

        Parameters
        ----------
        no_grad
            If True use torch.no_grad(), by default True.
        """
        training = True if self.training else False

        nc = contextlib.nullcontext()
        with torch.no_grad() if no_grad else nc:
            try:
                if training:
                    super().eval()
                yield
            finally:
                if training:
                    self.train()

    def migrate(self, device: Device) -> None:
        self.to(device=device)

    # def transform_inputs(self, *args, **kwargs):
    #     return

    def use_transform(self, transform: luz.TensorTransform) -> None:
        """Use transform.

        Parameters
        ----------
        transform
            Tensor transform applied to output during inference.
        """
        self.transform = transform



from __future__ import annotations
from typing import Any, Callable, Iterable, Optional, Union

from abc import ABC, abstractmethod
import argparse
import copy
import luz
import torch

__all__ = ["Learner"]

Device = Union[str, torch.device]
Criterion = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]


class Learner(ABC):
    def add_args(self, parser):
        raise NotImplementedError

    def parser(self):
        parser = argparse.ArgumentParser()
        self.add_args(parser)
        return parser

    def parse_args(self):
        parser = self.parser()
        self.args = parser.parse_args()

    @abstractmethod
    def model(self) -> luz.Model:
        """Return model to be trained.

        Returns
        -------
        luz.Model
            Model to be trained.
        """
        pass

    @abstractmethod
    def criterion(self) -> Criterion:
        pass

    @abstractmethod
    def optimizer(self, model: torch.nn.Module) -> torch.optim.Optimizer:
        pass

    @abstractmethod
    def fit_params(self) -> dict[str, Any]:
        """Return fit parameters.

        Parameters
        ----------
        train_dataset
            Training dataset used to learn a model.
        val_dataset
            Validation dataset.
        device
            Device to use for learning.

        Returns
        -------
        dict[str, Any]
            Dictionary of fit parameters.
        """
        pass

    def hyperparams(self, tuner: luz.Tuner) -> dict[str, Any]:
        """Specify hyperparameters for tuning.

        Parameters
        ----------
        tuner
            Hyperparameter tuner.

        Returns
        -------
        dict[str, Any]
            Dictionary of hyperparameters to be tuned.
            Each key is the hyperparameter name (used to access the sampled value),
            and values are tuning objects.

        Raises
        ------
        NotImplementedError
            [description]
        """
        raise NotImplementedError

    def scorer(self) -> luz.Scorer:
        raise NotImplementedError

    def tuner(self) -> luz.Tuner:
        raise NotImplementedError

    def handlers(self) -> Union[luz.Handler, Iterable[luz.Handler]]:
        return []

    def train_handlers(self) -> Union[luz.Handler, Iterable[luz.Handler]]:
        return self.handlers()

    def test_handlers(self) -> Union[luz.Handler, Iterable[luz.Handler]]:
        return self.handlers()

    def loggers(self) -> Union[luz.Logger, Iterable[luz.Logger]]:
        return luz.ConsoleLogger()

    def train_loggers(self) -> Union[luz.Logger, Iterable[luz.Logger]]:
        return self.loggers()

    def test_loggers(self) -> Union[luz.Logger, Iterable[luz.Logger]]:
        return self.loggers()

    def loader(self, dataset: luz.Dataset) -> torch.utils.data.DataLoader:
        return dataset.loader(
            batch_size=1,
            shuffle=True,
            num_workers=1,
            pin_memory=True,
        )

    def train_loader(self, dataset: luz.Dataset) -> torch.utils.data.DataLoader:
        return self.loader(dataset)

    def val_loader(self, dataset: luz.Dataset) -> torch.utils.data.DataLoader:
        return self.loader(dataset)

    def test_loader(self, dataset: luz.Dataset) -> torch.utils.data.DataLoader:
        return self.loader(dataset)

    def transform(self, train_dataset: luz.Dataset) -> luz.Transform:
        return None

    # def train_transform(self, train_dataset: luz.Dataset) -> luz.Transform:
    #     return self.transform(train_dataset)

    # def test_transform(self, train_dataset: luz.Dataset) -> luz.Transform:
    #     return self.transform(train_dataset)

    def learn(
        self,
        train_dataset: luz.Dataset,
        val_dataset: Optional[luz.Dataset] = None,
        device: Optional[Device] = "cpu",
    ) -> luz.Model:
        """Learn a model based on a given dataset.

        Parameters
        ----------
        train_dataset
            Training dataset used to learn a model.
        val_dataset
            Validation dataset, by default None.
        device
            Device to use for learning, by default "cpu".

        Returns
        -------
        luz.Model
            Learned model.
        """
        try:
            self.parse_args()
        except NotImplementedError:
            pass

        model = self.model()

        self._transform = self.transform(train_dataset)

        self.fit(model, train_dataset, val_dataset, device)

        return model

    def score(self, dataset, device: Optional[Device] = "cpu") -> luz.Score:
        """Learn a model and estimate its future performance based on a given dataset.

        Parameters
        ----------
        dataset
            Dataset used to learn and score a model.
        device
            Device to use for learning, by default "cpu".

        Returns
        -------
        luz.Score
            Learned model and estimated performance.
        """
        return self.scorer().score(self, dataset, device)

    def tune(
        self, dataset, device: Optional[Device] = "cpu"
    ) -> tuple[luz.Model, float, dict[str, Any]]:
        """Tune hyperparameters, learn a model with optimal hyperparameters,
        and estimate its future performance based on a given dataset.

        Parameters
        ----------
        dataset
            Dataset used to learn and score a model.
        device
            Device to use for learning, by default "cpu".

        Returns
        -------
        luz.Model
            Learned model with optimal hyperparameters.
        float
            Estimated future performance of learned model.
        dict[str, Any]
            Optimal hyperparameters.
        """
        tuner = self.tuner()
        self._tuner = tuner
        for exp in tuner.tune(**self.hyperparams(tuner)):
            self.hparams = exp
            model, score = tuner.score(self, dataset, device)

        return (
            tuner.best_model,
            tuner.best_score,
            tuner.best_hyperparameters,
        )

    def fit(
        self,
        model: luz.Module,
        train_dataset: luz.Dataset,
        val_dataset: Optional[luz.Dataset] = None,
        device: Optional[Device] = "cpu",
    ) -> None:
        """Train model.

        Parameters
        ----------
        model
            Model to be trained.
        train_dataset
            Training data.
        val_dataset
            Validation data, by default None.
        device
            Device to use for training, by default "cpu".
        """
        if self._transform is not None:
            train_loader = self.train_loader(
                train_dataset.use_transform(self._transform)
            )
            val_loader = self.val_loader(val_dataset.use_transform(self._transform))
        else:
            train_loader = self.train_loader(train_dataset)
            val_loader = self.val_loader(val_dataset)

        params = dict(start_epoch=1, stop_epoch=2, early_stopping=True, patience=5)
        params.update(self.fit_params())
        start_epoch = params["start_epoch"]
        stop_epoch = params["stop_epoch"]
        early_stopping = params["early_stopping"]
        patience = params["patience"]

        model.migrate(device)

        # NOTE: must come after migrate
        optimizer = self.optimizer(model=model)
        criterion = self.criterion()

        loggers = self.train_loggers()
        if isinstance(loggers, luz.Logger):
            # only 1 logger, so wrap it in a list
            loggers = [loggers]
        handlers = self.train_handlers()
        if isinstance(handlers, luz.Handler):
            # only 1 handler, so wrap it in a list
            handlers = [handlers]

        state = {}
        state.update(
            flag=luz.Flag.TRAINING,
            model=model,
            optimizer=optimizer,
            loader=train_loader,
            train_loader=train_loader,
            val_loader=val_loader,
            train_history=[],
            val_history=[],
            loggers=loggers,
        )
        if early_stopping:
            state.update(
                patience=patience, best_model=copy.deepcopy(model.state_dict())
            )

        luz.Event.TRAINING_STARTED(handlers, **state)

        for epoch in range(start_epoch, stop_epoch + 1):
            state.update(epoch=epoch)
            luz.Event.EPOCH_STARTED(handlers, **state)

            self.run_epoch(
                model, train_loader, device, handlers, criterion, state, optimizer
            )

            if val_dataset is not None:
                state.update(flag=luz.Flag.VALIDATING)

                # val_loss = self.validate(model, val_loader, device)

                with model.eval():
                    val_loss = self.run_epoch(
                        model,
                        val_loader,
                        device,
                        handlers,
                        criterion,
                        state,
                    )

                    state["val_history"].append(val_loss)

                    try:
                        # FIXME: replace 0.0 with self.delta_thresh?
                        if min(state["val_history"]) - val_loss < 0.0:
                            state["patience"] -= 1
                        else:
                            state.update(
                                patience=patience,
                                best_model=copy.deepcopy(model.state_dict()),
                            )
                    except ValueError:
                        pass

                # return val_loss

                # self.log(f"[Epoch {epoch}] Validation loss: {val_loss}.")

                if early_stopping and state["patience"] == 0:
                    # self.log(f"[Epoch {epoch}]: Stopping early.")
                    model.load_state_dict(state["best_model"])
                    break

                state.update(flag=luz.Flag.TRAINING)

        luz.Event.TRAINING_ENDED(handlers, **state)

    def test(
        self,
        model: luz.Module,
        dataset: luz.Dataset,
        device: Optional[Device] = "cpu",
    ) -> float:
        """Test model.

        Parameters
        ----------
        model
            Model to be tested.
        dataset
            Test data.
        device
            Device to use for testing, by default "cpu".
        """
        model.migrate(device)

        if self._transform is not None:
            loader = self.test_loader(dataset.use_transform(self._transform))
        else:
            loader = self.test_loader(dataset)

        criterion = self.criterion()

        loggers = self.test_loggers()
        if isinstance(loggers, luz.Logger):
            # only 1 logger, so wrap it in a list
            loggers = [loggers]
        handlers = self.test_handlers()
        if isinstance(handlers, luz.Handler):
            # only 1 handler, so wrap it in a list
            handlers = [handlers]

        state = {}
        state.update(
            flag=luz.Flag.TESTING,
            trainer=self,
            model=model,
            loader=loader,
            loggers=loggers,
        )
        luz.Event.TESTING_STARTED(handlers, **state)

        state.update(epoch=1)
        luz.Event.EPOCH_STARTED(handlers, **state)

        test_loss = self.run_epoch(model, loader, device, handlers, criterion, state)

        luz.Event.TESTING_ENDED(handlers, **state)

        return test_loss

    def run_epoch(
        self,
        model: luz.Module,
        loader: torch.utils.data.DataLoader,
        device: Device,
        handlers,
        criterion: Criterion,
        state,
        optimizer: Optional[torch.optim.Optimizer] = None,
    ) -> float:
        running_loss = 0.0

        for i, batch in enumerate(loader):
            data = model.get_input(batch)
            target = model.get_target(batch)

            state.update(ind=i, data=data, target=target)
            luz.Event.BATCH_STARTED(handlers, **state)

            # migrate the input and target tensors to the appropriate device
            data, target = data.to(device), target.to(device)

            if state["flag"] == luz.Flag.TRAINING:
                output, batch_loss = model.run_train_batch(
                    data, target, criterion, optimizer
                )
            elif state["flag"] == luz.Flag.VALIDATING:
                with model.eval():
                    output, batch_loss = model.run_validate_batch(
                        data, target, criterion
                    )
            elif state["flag"] == luz.Flag.TESTING:
                with model.eval():
                    output, batch_loss = model.run_test_batch(data, target, criterion)

            running_loss += batch_loss.item()

            state.update(output=output, loss=batch_loss)

            # from https://coolnesss.github.io/2019-02-05/pytorch-gotchas
            # All of the variables defined above are now out of scope!
            # On CPU, they are already deallocated.
            # On GPU, they will be deallocated soon.

            # Make sure deallocation has taken place
            if torch.cuda.is_available():
                torch.cuda.synchronize()

            luz.Event.BATCH_ENDED(handlers, **state)

        loss = running_loss / len(loader)

        if state["flag"] == luz.Flag.TRAINING:
            state["train_history"].append(loss)

        luz.Event.EPOCH_ENDED(handlers, **state)

        return loss

    def state_dict(self):
        return {
            "loss": self.params.loss,
            "optimizer": self.params.optimizer,
            "start_epoch": self.params.start_epoch,
            "stop_epoch": self.params.stop_epoch,
            "early_stopping": self.params.early_stopping,
            "patience": self.params.patience,
            # "loader_kwargs": self._fit_params["loader_kwargs"],
            "handlers": self.handlers,
            "loggers": self.params.loggers,
            # "_state": self._state,
        }

    def load_state_dict(self, state_dict):
        self.loss = state_dict["loss"]
        self.optimizer = state_dict["optimizer"]
        self.start_epoch = state_dict["start_epoch"]
        self.stop_epoch = state_dict["stop_epoch"]
        self.early_stopping = state_dict["early_stopping"]
        self.patience = state_dict["patience"]
        # self._fit_params["loader_kwargs"] = state_dict["loader_kwargs"]
        self.handlers = state_dict["handlers"]
        self.loggers = state_dict["loggers"]
        # self._state = state_dict["_state"]



import enum

__all__ = ["Event"]


class Event(enum.Enum):
    BATCH_STARTED = "BATCH STARTED"
    BATCH_ENDED = "BATCH ENDED"
    EPOCH_STARTED = "EPOCH STARTED"
    EPOCH_ENDED = "EPOCH ENDED"
    TESTING_STARTED = "TESTING STARTED"
    TESTING_ENDED = "TESTING ENDED"
    TRAINING_STARTED = "TRAINING STARTED"
    TRAINING_ENDED = "TRAINING ENDED"
    VALIDATING_STARTED = "VALIDATING STARTED"
    VALIDATING_ENDED = "VALIDATING ENDED"

    def __call__(self, handlers, **kwargs):
        for h in handlers:
            getattr(h, self.name.lower())(**kwargs)



"""

Contains callback objects which perform various functions during the training process.

"""

# FIXME: Write a test for every handler here (many are likely broken)
# FIXME: Type annotate every handler here

from __future__ import annotations
from typing import Any, Iterable, Optional, Union

import datetime
import luz
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import numpy as np
import os
import pathlib
import torch

__all__ = [
    "Handler",
    "Accuracy",
    "ActualVsPredicted",
    # "Checkpoint",
    # "DataAndFit",
    "DurbinWatson",
    "FBeta",
    "Loss",
    "PlotHistory",
    "Progress",
    # "RVP",
    "Timer",
]


class Handler:
    def batch_started(self, **kwargs: Any):
        pass

    def batch_ended(self, **kwargs: Any):
        pass

    def epoch_started(self, **kwargs: Any):
        pass

    def epoch_ended(self, **kwargs: Any):
        pass

    def testing_started(self, **kwargs: Any):
        pass

    def testing_ended(self, **kwargs: Any):
        pass

    def training_started(self, **kwargs: Any):
        pass

    def training_ended(self, **kwargs: Any):
        pass

    def log(self, loggers, msg: str) -> None:
        for logger in loggers:
            logger.log(msg)

    # def train_batch_started(self, **kwargs: Any):
    #         pass

    #     def train_batch_ended(self, **kwargs: Any):
    #         pass

    #     def test_batch_started(self, **kwargs: Any):
    #         pass

    #     def test_batch_ended(self, **kwargs: Any):
    #         pass

    #     def val_batch_started(self, **kwargs: Any):
    #         pass

    #     def val_batch_ended(self, **kwargs: Any):
    #         pass

    #     def train_epoch_started(self, **kwargs: Any):
    #         pass

    #     def train_epoch_ended(self, **kwargs: Any):
    #         pass

    #     def test_epoch_started(self, **kwargs: Any):
    #         pass

    #     def test_epoch_ended(self, **kwargs: Any):
    #         pass

    #     def val_epoch_started(self, **kwargs: Any):
    #         pass

    #     def val_epoch_ended(self, **kwargs: Any):
    #         pass

    #     def testing_started(self, **kwargs: Any):
    #         pass

    #     def testing_ended(self, **kwargs: Any):
    #         pass

    #     def training_started(self, **kwargs: Any):
    #         pass

    #     def training_ended(self, **kwargs: Any):
    #         pass

    def validating_started(self, **kwargs: Any):
        pass

    def validating_ended(self, **kwargs: Any):
        pass


class Accuracy(Handler):
    def __init__(self) -> None:
        self.correct = 0
        self.total = 0

    def epoch_started(self, **kwargs: Any) -> None:
        """Compute on epoch start."""
        self.correct = 0
        self.total = 0

    def batch_ended(
        self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any
    ) -> None:
        """Compute on batch end.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,C)`
        target
            Target tensor. One-hot encoded.
            Shape: :math:`(N,C)`
        """
        predicted = torch.argmax(torch.softmax(output, dim=1), dim=1)
        correct = torch.argmax(target, dim=1)

        self.correct += (predicted == correct).sum().item()
        self.total += target.size(0)

    def epoch_ended(self, model, epoch: int, loggers, **kwargs: Any) -> None:
        acc = self.correct / self.total
        s = f"[Epoch {epoch}] Classification accuracy: {acc}"
        self.log(loggers, s)


class ActualVsPredicted(Handler):
    def __init__(self, filepath: Optional[Union[str, pathlib.Path]] = None) -> None:
        """Plot actual labels vs. predicted labels.

        Parameters
        ----------
        filepath
            Path to save plot if not None, by default None.
        """
        self.filepath = filepath

    def batch_ended(
        self, output: torch.Tensor, target: torch.Tensor, flag: luz.Flag, **kwargs: Any
    ) -> None:
        """Compute on batch end.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,)`
        target
            Target tensor.
            Shape: :math:`(N,)`
        """
        if flag == luz.Flag.TESTING:
            x = output.detach().reshape(-1).numpy()
            y = target.detach().reshape(-1).numpy()
            self.ax.scatter(x, y, color="black")

    def testing_started(self, **kwargs: Any) -> None:
        """Compute on testing start."""
        self.fig, self.ax = plt.subplots()

    def testing_ended(self, **kwargs: Any) -> None:
        """Compute on testing end."""
        x_min, x_max = self.ax.get_xlim()
        y_min, y_max = self.ax.get_ylim()
        plot_min = min(x_min, y_min)
        plot_max = max(x_max, y_max)
        self.ax.set_xlim(plot_min, plot_max)
        line = mlines.Line2D([0, 1], [0, 1], color="red")
        line.set_transform(self.ax.transAxes)
        self.ax.add_line(line)
        plt.ylabel("Predicted")
        plt.xlabel("Actual")
        plt.title("Predicted vs. actual")
        if self.filepath is not None:
            plt.savefig(luz.expand_path(self.filepath))

        plt.show()


class Checkpoint(Handler):
    def __init__(self, save_interval, model_name, save_dir=None):
        if save_interval < 0:
            raise ValueError(
                "The number of epochs between checkpoints must be a positive integer."
            )

        self.model_name = model_name
        self.save_dir = luz.utils.expand_path(path=save_dir)

        luz.mkdir_safe(self.save_dir)

        self.save_interval = save_interval

    # # FIXME: Fix this
    # def state_dict(self):
    #     return {}

    def epoch_ended(self, model, epoch, **kwargs):
        if epoch % self.save_interval == 0:
            save_path = os.path.join(
                self.save_dir, f"{model.model_name}_{epoch}.pth.tar"
            )
            torch.save(obj=model.state_dict(), f=save_path)


class DataAndFit(Handler):
    def __init__(self):
        self.xs = []
        self.actual = []
        self.predicted = []

    def update(self, x, y, predicted):
        self.xs.append(x.item())
        self.actual.append(y.item())
        self.predicted.append(predicted.item())

    def compute(self):
        plt.scatter(self.xs, self.actual, label="Actual data")
        plt.scatter(self.xs, self.predicted, label="Fit function")
        plt.xlabel("Input")
        plt.ylabel("Output")
        plt.title("Data and fit")
        plt.show()


class DurbinWatson(Handler):
    def __init__(self):
        self.numerator = 0
        self.denominator = 0
        self.last_residual = None

    def batch_ended(self, x, y, output, **kwargs):
        residual = (y - output).item()
        if self.last_residual is not None:
            self.numerator += (residual - self.last_residual) ** 2
        self.last_residual = residual
        self.denominator += residual ** 2

    def epoch_ended(self, model, loggers, **kwargs):
        self.log(loggers, f"DW: {self.numerator / self.denominator}")


class FBeta(Handler):
    def __init__(self, beta: float) -> None:
        self.beta = beta

        self.true_positive = 0
        self.predicted_positive = 0
        self.actual_positive = 0

    def epoch_started(self, **kwargs: Any) -> None:
        """Compute on epoch start."""
        self.true_positive = 0
        self.predicted_positive = 0
        self.actual_positive = 0

    def batch_ended(
        self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any
    ) -> None:
        """Compute on batch end.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,2)`
        target
            Target tensor. One-hot encoded.
            Shape: :math:`(N,2)`
        """
        predicted = torch.argmax(torch.softmax(output, dim=1), dim=1)
        correct = torch.argmax(target, dim=1)

        self.true_positive += correct[predicted.nonzero(as_tuple=False)].sum().item()
        self.predicted_positive += predicted.sum().item()
        self.actual_positive += correct.sum().item()

    def epoch_ended(self, model, epoch: int, loggers, **kwargs: Any) -> None:
        try:
            precision = self.true_positive / self.predicted_positive
            recall = self.true_positive / self.actual_positive
            F = (
                (1 + self.beta ** 2)
                * precision
                * recall
                / (precision * (self.beta ** 2) + recall)
            )
        except ZeroDivisionError:
            F = 1

        s = f"[Epoch {epoch}] F-score: {F}"
        self.log(loggers, s)


class Loss(Handler):
    def __init__(self, print_interval: Optional[float] = 0.25) -> None:
        """Calculate running loss throughout each epoch.

        Parameters
        ----------
        print_interval
            Fraction of epoch after which the running loss is printed, by default 0.25.

        Raises
        ------
        ValueError
        """
        if print_interval <= 0 or print_interval > 1:
            raise ValueError(
                "Print interval must be a positive number between 0.0 and 1.0."
            )

        self.print_interval = print_interval

    def epoch_started(self, **kwargs: Any) -> None:
        self.running_loss = 0

    def batch_ended(
        self,
        flag: luz.Flag,
        loader: torch.utils.data.DataLoader,
        epoch: int,
        loss: torch.Tensor,
        ind: int,
        loggers,
        **kwargs: Any,
    ) -> None:
        """Execute at end of batch.

        Parameters
        ----------
        flag
            Training flag.
        loader
            Data loader.
        epoch
            Epoch number.
        loss
            Last computed loss.
        ind
            Data index.
        **kwargs
            Superfluous kwargs.
        """
        if flag == luz.Flag.TRAINING:
            # NOTE: it's very important to add loss.item()
            # (as opposed to loss) to avoid a memory leak!
            self.running_loss += loss.item()
            num_batches = len(loader)
            batch_interval = max(1, round(num_batches * self.print_interval))
            cur = ind + 1
            if cur % batch_interval == 0 or cur == num_batches:
                avg_loss = self.running_loss / num_batches
                self.log(loggers, f"[Epoch {epoch}] Running average loss: {avg_loss}.")


class PlotHistory(Handler):
    def __init__(
        self,
        show_plot: Optional[bool] = True,
        save_filepath: Optional[str] = None,
        reduction: Optional[str] = "mean",
    ) -> None:
        self.show_plot = show_plot
        self.save_filepath = save_filepath
        self.reduction = reduction

    def training_ended(
        self,
        train_history: Iterable[float],
        val_history: Iterable[float],
        train_loader: torch.utils.data.DataLoader,
        val_loader: torch.utils.data.DataLoader,
        **kwargs: Any,
    ) -> None:
        x = range(1, len(train_history) + 1)

        fig, ax1 = plt.subplots()
        ax1.set_xlabel("Epoch")

        if self.reduction == "mean":
            # divide each epoch loss (which is the sum of batch averages)
            # by the number of batches to estimate average epoch loss
            y1 = np.array(train_history) / len(train_loader)

        (line1,) = ax1.plot(x, y1, color="tab:blue")

        if len(val_history) > 0:
            ax2 = ax1.twinx()

            if self.reduction == "mean":
                # divide each epoch loss (which is the sum of batch averages)
                # by the number of batches to estimate average epoch loss
                y2 = np.array(val_history) / len(val_loader)

            (line2,) = ax2.plot(x, y2, color="tab:orange")

            plt.title("Loss history")
            plt.legend((line1, line2), ("Training loss", "Validation loss"))
        else:
            plt.title("Loss history")
            plt.legend((line1,), ("Training loss",))

        fig.tight_layout()

        if self.save_filepath is not None:
            plt.savefig(self.save_filepath)

        if self.show_plot:
            plt.show()
        else:
            plt.close(fig)


class Progress(Handler):
    def __init__(
        self, print_interval: Optional[float] = 0.25, bar_length: Optional[int] = 30
    ) -> None:
        if print_interval <= 0 or print_interval > 1:
            raise ValueError(
                "Print interval must be a positive number between 0.0 and 1.0."
            )

        self.bar_length = bar_length
        self.print_interval = print_interval

    def batch_ended(
        self,
        loader: torch.utils.data.DataLoader,
        epoch: int,
        flag: luz.Flag,
        ind: int,
        loggers,
        **kwargs: Any,
    ) -> None:
        if flag == luz.Flag.TRAINING:
            progress = (ind + 1) / len(loader)

            if int(progress / self.print_interval) > self._last or progress == 1:
                num_eqs = int(progress * self.bar_length)
                num_spaces = self.bar_length - num_eqs
                bar = "=" * num_eqs + ">" + " " * num_spaces
                self.log(loggers, f"[Epoch {epoch}]: [{bar}] {int(100*progress)}%")
                self._last += 1

    def epoch_started(self, loggers, **kwargs: Any) -> None:
        self._last = 0

    def training_started(self, loggers, **kwargs: Any) -> None:
        self.log(loggers, "Training started.")

    def training_ended(self, loggers, **kwargs: Any) -> None:
        self.log(loggers, "Training ended.")

    def testing_started(self, loggers, **kwargs: Any) -> None:
        self.log(loggers, "Testing started.")

    def testing_ended(self, loggers, **kwargs: Any) -> None:
        self.log(loggers, "Testing complete.")


class RVP(Handler):
    def __init__(self) -> None:
        self.predicted = []
        self.residual = []

    def batch_ended(
        self, x: torch.Tensor, y: torch.Tensor, output: torch.Tensor, **kwargs: Any
    ) -> None:
        self.predicted.append(output.item())
        self.residual.append((y - output).item())

    def training_ended(self, **kwargs: Any) -> None:
        self.fig, self.ax = plt.subplots()
        self.ax.scatter(self.predicted, self.residual, c="black")
        mmax = max([abs(max(self.residual)), abs(min(self.residual))])
        plt.ylim([-mmax, mmax])
        plt.axhline(y=0.0, color="r", linestyle="-")
        plt.ylabel("Residual")
        plt.xlabel("Predicted")
        plt.title("Residual vs. predicted")
        plt.show()


class Timer(Handler):
    def __init__(self) -> None:
        self.start_time = 0
        self.end_time = 0

    def epoch_started(self, **kwargs: Any) -> None:
        self.start_time = datetime.datetime.now().replace(microsecond=0)

    def epoch_ended(self, epoch: int, flag: luz.Flag, loggers, **kwargs: Any) -> None:
        self.end_time = datetime.datetime.now().replace(microsecond=0)

        if flag == luz.Flag.TRAINING:
            mode = "train"
        elif flag == luz.Flag.VALIDATING:
            mode = "validate"
        else:
            mode = "test"

        self.log(
            loggers, f"[Epoch {epoch}]: {self.end_time-self.start_time} to {mode}."
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['luz']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3.2,<4.0.0', 'torch>=1.7.0,<2.0.0']

setup_kwargs = {
    'name': 'luz',
    'version': '8.0.0',
    'description': 'Framework for rapid research and development of machine learning projects using PyTorch.',
    'long_description': "==============\nLuz Module\n==============\n\n.. image:: https://codecov.io/gh/kijanac/luz/branch/master/graph/badge.svg\n  :target: https://codecov.io/gh/kijanac/luz\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github\n\n**Framework for rapid research and development of machine learning projects using PyTorch.**\n\nLonger description coming soon!\n\nImportant features:\n\n#. Reduced boilerplate and simple specifications of complex workflows\n#. Easy and flexible customization of Learner by overriding methods\n#. Built-in scoring algorithms like holdout and cross validation\n#. Straightforward hyperparameter tuning with built-in tuning algorithms like random search and grid search\n\nBasic usage: custom modules with luz.Module functionality. Simply define your model, inherit from luz.Module, and use model.train/model.test.\n\nNext level: simplified training algorithms with transforms and handlers.\n\nNext level: straightforward hyperparameter tuning with various scoring mechanisms.\n\n#. Data preparation scheme.\n\n  #. Transforms: Functions for conditionining the training process which are invertible and whose inverse is applied to the resulting predictor.\n  #. Preprocessors: Functions for altering the prediction task which are not necessarily invertible.\n\n#. Model.\n#. Training scheme.\n#. Overall learning algorithm.\n\n  #. Hyperparameter selection.\n\n#. Unified development interface through Learner object. Simply inherit luz.Learner, define the model, loader, and param functions, and you're good to go. Add a hyperparams function to enable tuning and make tuned parameters accessible in the model and param functions.\n\n---------------\nGetting Started\n---------------\n\nPrerequisites\n-------------\n\nInstalling\n----------\n\nTo install, open a shell terminal and run::\n\n`conda create -n luz -c conda-forge -c pytorch -c kijana luz`\n\n----------\nVersioning\n----------\n\n-------\nAuthors\n-------\n\nKi-Jana Carter\n\n-------\nLicense\n-------\nThis project is licensed under the MIT License - see the LICENSE file for details.\n",
    'author': 'Ki-Jana Carter',
    'author_email': 'kijana@mit.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kijanac/luz',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
